require 'utils'

function trim(s)
	return s:gsub("^%s+", ""):gsub("%s+$", "")
end

-- only declarations, no #includes
function preprocess(fn)
	local f = io.open(fn)
	local str = f:read("*a")
	str = str:gsub("%#include .-\n", ""):gsub("//.-\n", ""):gsub("%/%*.-%*%/", "")	
	return str
end

function match(fn, pat)
	local fid = io.open(fn)
	if not fid then return end
	local ret = fid:read("*a"):match(pat)
	fid:close()
	return ret
end

local update = false
local java = false
local package = false
for i = 1, #arg do
	if arg[i] == "--update" then update = true end
	if arg[i] == "--java" then java = true end
	if arg[i] == "--package" then package = true end	
end

-- Update version number
if update then
	local fid = io.open("../src/systemic.h", "r")
	local all = fid:read("*a")
	local ver = 0
	all = all:gsub("%#define VERSION ([%de%+%-%.]+)", function(cap) 
		ver = tonumber(cap) + 1e-4
		return string.format("#define VERSION %.4f", ver)
	end)
	fid:close()
	fid = io.open("../src/systemic.h", "w")
	fid:write(all)
	fid:close()
end

if package then
	local ver = match("../src/systemic.h", "%#define VERSION ([%de%+%-%.]+)")
	local fid = io.open("../../package.sh", "w")
	fid:write(string.format([[make -f Makefile.osx clean cleanreqs
	make -f Makefile.osx reqs	
	make -f Makefile.osx all_in_one
	make -f Makefile.osx clean cleanreqs
	export COPY_EXTENDED_ATTRIBUTES_DISABLE=true
	export COPYFILE_DISABLE=true
	tar -cvjhf ~/Downloads/systemic_%s.tar.bz2 --exclude '*.class' --exclude '*.svn' --exclude '.*' --exclude '*.o' --exclude='._*' --exclude 'private' --exclude 'test.c' ./*
	tar -cvjhf ~/Downloads/systemic_%s_mac.tar.bz2 --exclude '*.class' --exclude '*.svn' --exclude='._*' --exclude '.*' --exclude '*.o' --exclude 'private' --exclude 'test.c' ./mac]], ver:gsub("%.", ""), tostring(ver):gsub("%.", "")))
	fid:close()
end

-- Parses the systemic.h header and creates the defs.lua file
local defs = io.open("defs.lua", "w")
local javaconsts
if java then
	javaconsts = io.open("../../SystemicGui/src/repl/Defs.java", "w")
end

local all = buf.buffer():append(preprocess("../src/utils.h"), preprocess("../src/systemic.h"), preprocess("../src/kernel.h"), preprocess("../src/integration.h"), preprocess("../src/kl.h"), preprocess("../src/mcmc.h"), preprocess("../src/bootstrap.h"), preprocess("../src/periodogram.h")):tostring()

-- Digest declarations
defs:write[==[
-- Automatically generated by parse.lua
ffi = require 'ffi'

_sysc = {}
]==]

if java then 
	javaconsts:write[==[
package repl;
public class Defs {
]==]
end

local decls = buf.buffer()
-- Simple numbers
for lab, num in all:gmatch("%#define ([%w_]+) ([%de%+%-%.]+)") do

	defs:write(string.format("_sysc.%s = %s\n", lab, num))
	
	if java then
		local lnum = tonumber(num)
		if math.abs(lnum - math.floor(lnum)) < 1e-8 and math.abs(lnum) < 1e6 then
			javaconsts:write(string.format("\tpublic static int %s = %d;\n", lab, lnum))
		else
			javaconsts:write(string.format("\tpublic static double %s = %s;\n", lab, num))
		end
	end
end
-- Flags
for lab, num in all:gmatch("%#define ([%w_]+) %(1 << (%d+)%)") do
	defs:write(string.format("_sysc.%s = %d\n", lab, 2^tonumber(num)))
	if java then
		javaconsts:write(string.format("\tpublic static int %s = %d;\n", lab, 2^tonumber(num)))
	end
end

if java then
	javaconsts:write("}\n");
	javaconsts:close()
end
-- Preprocess!
local fid = io.open(".tmp.h", "w")
fid:write(all)
fid:close()

local digest = buf.buffer()
digest:append[[
typedef void FILE;

typedef void gsl_rng;
typedef void gsl_vector_view;
void free(void*);
FILE * fopen ( const char * filename, const char * mode );
int fclose ( FILE * stream );
size_t fwrite ( const void * ptr, size_t size, size_t count, FILE * stream );
int fputs ( const char * str, FILE * stream );


typedef struct
     {
       size_t size;
       double * data;
     } gsl_block;

typedef struct
     {
       size_t size;
       double * data;
     } gsl_block_int;

   typedef struct
     {
       size_t size;
       size_t stride;
       double * data;
       gsl_block * block;
       int owner;
     } gsl_vector;

   typedef struct
     {
       size_t size;
       size_t stride;
       int * data;
       gsl_block_int * block;
       int owner;
     } gsl_vector_int;

   typedef struct
     {
       size_t size1;
       size_t size2;
       size_t tda;
       double * data;
       gsl_block * block;
       int owner;
     } gsl_matrix;

typedef struct
     {
       size_t size1;
       size_t size2;
       size_t tda;
       int * data;
       gsl_block_int * block;
       int owner;
     } gsl_matrix_int;

gsl_matrix * gsl_matrix_alloc (size_t n1, size_t n2);
void gsl_matrix_free (gsl_matrix * m);
int gsl_matrix_fwrite (FILE * stream, const gsl_matrix * m);
int gsl_matrix_fread (FILE * stream, gsl_matrix * m);
int gsl_matrix_fprintf (FILE * stream, const gsl_matrix * m, const char * format);
int gsl_matrix_fscanf (FILE * stream, gsl_matrix * m);
double gsl_matrix_get (const gsl_matrix * m, size_t i, size_t j);
void gsl_matrix_set (gsl_matrix * m, size_t i, size_t j, double x);

]]
digest:append(ex.capture("/usr/local/bin/cpp -P -E -w .tmp.h"))

defs:write(string.format("ffi.cdef[===[\n%s\n]===]\n", digest:tostring()))
defs:write[[
require 'utils'
if ex.exists("libsystemic.dylib") then
	_sys = ffi.load("./libsystemic.dylib")
else
	_sys = ffi.load("./libsystemic.so")
end
]]

defs:close()